#!/usr/bin/env python3

'''
This command creates the packages and packages.gz file
for all *.ipk files in the folder.

Usage: ./make_package_index.py [folder]
Operations are done by default in the working directory
you can specify the folder however to change the folder
that is worked in.
'''

# Even though this script is not on GitHub (as it's just
# a test for a project that will probably never get anything
# remotely serious for me). I hereby grant you permission
# to take this script and use it for your own purposes 
# wherever you want with any license.
# - Linus K. <linus@cosmos-ink.net>

import gzip
import hashlib
import io
import os
import sys
import tarfile


# Keep a trailing comma if only one entry!
PACKAGES_FIELDS_HEAD = 'Package', 'Version',
PACKAGES_FIELDS_GENERATED = 'Filename', 'Size', 'SHA256sum',
PACKAGES_FIELDS_TAIL = 'Description',
PACKAGES_FIELDS_BLOCKED = 'Source', 'SourceName',


def find(archive: tarfile.TarFile, normalized_path: str) -> tarfile.TarInfo:
    for member in archive.getmembers():
        if os.path.normpath(member.path) == normalized_path:
            return member
    return None

def get_control_content(ipkg_filename: str) -> str:
    ''' Get full content of "filename/control.tar.gz/control" as utf-8 decoded str. '''
    with open(ipkg_filename, 'rb') as ipkg_archive_fp:
        ipkg_archive: tarfile.TarFile = tarfile.open(fileobj=ipkg_archive_fp, mode='r:gz')
        with ipkg_archive.extractfile(find(ipkg_archive, 'control.tar.gz')) as control_archive_fp:
            control_archive: tarfile.TarFile = tarfile.open(fileobj=control_archive_fp, mode='r:gz')
            with control_archive.extractfile(find(control_archive, 'control')) as control_file_fp:
                return control_file_fp.read().decode('utf-8')


def get_control_dict(ipkg_filename: str) -> dict:
    ''' Same as get_control_content but also parses it into a dict afterwards. '''
    result = {}

    current_key = None
    current_value = None
    control_content = get_control_content(ipkg_filename)
    for line in control_content.split('\n'):
        if current_key is not None and current_value is not None:
            result[current_key] = current_value

        if not line or line.startswith('#'):
            # Ignore comments and empty lines
            continue
        
        if line.startswith(' '):
            if current_key is None or current_value is None:
                raise RuntimeError('Got continuing value as first entry: %s' % repr(line))
            # Value of last entry continues. Example:
            # Description: Hello this is a multiline
            #   description. This second line and
            #   all following have a space in front
            #   and still are the value for Description.
            current_value += '\n' + line[1:].strip()
        else:
            #Key value pair (or start of value)
            if ':' not in line:
                raise RuntimeError('Line has no colon: %s' % repr(line))
            col_index = line.index(':')
            current_key = line[:col_index]
            current_value = line[col_index+1:].strip()
    
    if current_key is not None and current_value is not None:
        result[current_key] = current_value
    return result


def sha256sum(filename: str) -> str:
    ''' Some method to buffer the given file and return the sha256 hash. '''
    # https://nitratine.net/blog/post/how-to-hash-files-in-python/
    file = ".\myfile.txt" # Location of the file (can be set a different way)
    BLOCK_SIZE = 65536 # The size of each read from the file

    file_hash = hashlib.sha256() # Create the hash object, can use something other than `.sha256()` if you wish
    with open(filename, 'rb') as f: # Open the file to read it's bytes
        fb = f.read(BLOCK_SIZE) # Read from the file. Take in the amount declared above
        while len(fb) > 0: # While there is still data being read from the file
            file_hash.update(fb) # Update the hash
            fb = f.read(BLOCK_SIZE) # Read the next block from the file

    return file_hash.hexdigest() # Get the hexadecimal digest of the hash


def write_field(fp, key: str, value):
    ''' Used by write() '''
    if not isinstance(value, str):
        value = str(value)
    
    if '\n' in value:
        lines = value.split('\n')
        fp.write(f'{key}: {lines[0]}\n')
        for line in lines[1:]:
            fp.write(f' {line}\n')
    else:
        fp.write(f'{key}: {value}\n')


def write_control_dict(fp, control_dict: dict, filename: str):
    # Write leading fields in order first (e.g. Package, ...)
    for fieldname in PACKAGES_FIELDS_HEAD:
        if fieldname in control_dict:
            write_field(fp, fieldname, control_dict[fieldname])
    
    # Write fields that are neither heading nor trailing (e.g. Architecture, ...)
    for fieldname in control_dict:
        # Write all remaing fields that are not in any PACKAGE_FIELDS_* sets
        if fieldname not in PACKAGES_FIELDS_HEAD and fieldname not in PACKAGES_FIELDS_TAIL \
            and fieldname not in PACKAGES_FIELDS_GENERATED and fieldname not in PACKAGES_FIELDS_BLOCKED:
            write_field(fp, fieldname, control_dict[fieldname])
    
    # Generated fields that should be in PACKAGES_FIELDS_GENERATED
    write_field(fp, 'Filename', filename)
    write_field(fp, 'Size', os.path.getsize(filename))
    write_field(fp, 'SHA256Sum', sha256sum(filename))

    # Write trailing fields in order last (e.g. Description, ...)
    for fieldname in PACKAGES_FIELDS_TAIL:
        if fieldname in control_dict:
            write_field(fp, fieldname, control_dict[fieldname])

    fp.write('\n') # End with empty line to mark end of this package


def version_parts(version_string):
    num = ''
    for char in version_string:
        if char in '0123456789':
            num += char
        else:
            yield int(num)
            num = ''
    if num != '':
        yield int(num)

def most_recent(*version_strings):
    most_recent = None
    most_recent_parts = None
    for version_string in version_strings:
        parts = tuple(version_parts(version_string))
        if most_recent is None:
            most_recent = version_string
            most_recent_parts = parts
        else:
            for part_mr, part_vs in zip(most_recent_parts, parts):
                if part_vs > part_mr:
                    most_recent = version_string
                    most_recent_parts = parts
                    break
                elif part_vs < part_mr:
                    # This is not more recent
                    break
            else:
                if len(parts) > len(most_recent_parts):
                    most_recent = version_string
                    most_recent_parts = parts
    
    return most_recent

def write_packages_files(filenames):
    package_control_dicts = {} # name: (filename, control_dict)
    for filename in filenames:
        control_dict = get_control_dict(filename)
        name = control_dict['Package']
        version = control_dict['Version']
        if name not in package_control_dicts \
            or most_recent(version, package_control_dicts[name][1]['Version']) == version:
            # Either non existant or this version is more recent
            package_control_dicts[name] = (filename, control_dict)

    with io.StringIO() as buffer:
        with open('Packages', 'w') as packages_fp:
            with gzip.open('Packages.gz', 'wt') as packages_gz_fp:
                for (filename, control_dict) in package_control_dicts.values():
                    # Write to buffer
                    write_control_dict(buffer, control_dict, filename)

                    # Read out buffer
                    buffer.seek(0)
                    buffer_content = buffer.read()
                    buffer.seek(0)
                    buffer.truncate()

                    # Write buffer to both files
                    packages_fp.write(buffer_content)
                    packages_gz_fp.write(buffer_content)


if __name__ == '__main__':
    if len(sys.argv) > 1: # Has postional argument [folder] ?
        os.chdir(sys.argv[1]) # Work in another directory
    
    filenames = list(filter(lambda filename: filename.endswith('.ipk'), os.listdir()))
    filenames.sort() # Sort alphabetically

    write_packages_files(filenames)
